# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Pedro Algarvio (pedro@algarvio.me)`


    ==========================================
    PyLint Extended String Formatting Checkers
    ==========================================

    Proper string formatting PyLint checker
'''

# Import Python libs
from __future__ import absolute_import
import re
import sys
import tokenize

# Import PyLint libs
try:
    # >= pylint 1.0
    import astroid
except ImportError:  # pylint < 1.0
    from logilab import astng as astroid  # pylint: disable=no-name-in-module
from saltpylint.checkers import BaseChecker, utils
from pylint.checkers import BaseTokenChecker

try:
    # >= pylint 1.0
    from pylint.interfaces import IAstroidChecker
except ImportError:  # < pylint 1.0
    from pylint.interfaces import IASTNGChecker as IAstroidChecker  # pylint: disable=no-name-in-module

from pylint.interfaces import ITokenChecker, IRawChecker
from astroid.exceptions import InferenceError
try:
    from astroid.exceptions import NameInferenceError
except ImportError:
    class NameInferenceError(Exception):
        pass

# Import 3rd-party libs
import six


STRING_FORMAT_MSGS = {
    'W1320': ('String format call with un-indexed curly braces: %r',
              'un-indexed-curly-braces-warning',
              'Under python 2.6 the curly braces on a \'string.format()\' '
              'call MUST be indexed.'),
    'E1320': ('String format call with un-indexed curly braces: %r',
              'un-indexed-curly-braces-error',
              'Under python 2.6 the curly braces on a \'string.format()\' '
              'call MUST be indexed.'),
    'W1321': ('String substitution used instead of string formattting on: %r',
              'string-substitution-usage-warning',
              'String substitution used instead of string formattting'),
    'E1321': ('String substitution used instead of string formattting on: %r',
              'string-substitution-usage-error',
              'String substitution used instead of string formattting'),
    'E1322': ('Repr flag (!r) used in string: %r',
              'repr-flag-used-in-string',
              'Repr flag (!r) used in string'),
    'E1323': ('String formatting used in logging: %r',
              'str-format-in-logging',
              'String formatting used in logging'),
}

BAD_FORMATTING_SLOT = re.compile(r'(\{![\w]{1}\}|\{\})')


class StringCurlyBracesFormatIndexChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'string'
    msgs = STRING_FORMAT_MSGS
    priority = -1

    options = (('un-indexed-curly-braces-always-error',
                {'default': 1, 'type': 'yn', 'metavar': '<y_or_n>',
                 'help': 'Force un-indexed curly braces on a '
                         '\'string.format()\' call to always be an error.'}
                ),
                ('enforce-string-formatting-over-substitution',
                 {'default': 1, 'type': 'yn', 'metavar': '<y_or_n>',
                  'help': 'Enforce string formatting over string substitution'}
                 ),
                ('string-substitutions-usage-is-an-error',
                 {'default': 1, 'type': 'yn', 'metavar': '<y_or_n>',
                  'help': 'Force string substitution usage on strings '
                          'to always be an error.'}
                 ),
               )

    @utils.check_messages(*(STRING_FORMAT_MSGS.keys()))
    def visit_binop(self, node):
        if not self.config.enforce_string_formatting_over_substitution:
            return

        if node.op != '%':
            return

        if not (isinstance(node.left, astroid.Const) and
                isinstance(node.left.value, six.string_types)):
            return

        try:
            required_keys, required_num_args = utils.parse_format_string(node.left.value)[:2]
        except (utils.UnsupportedFormatCharacter, utils.IncompleteFormatString):
            # This is handled elsewere
            return

        if required_keys or required_num_args:
            if self.config.string_substitutions_usage_is_an_error:
                msgid = 'E1321'
            else:
                msgid = 'W1321'
            self.add_message(
                msgid, node=node.left, args=node.left.value
            )

        if '!r}' in node.left.value:
            self.add_message(
                'E1322', node=node.left, args=node.left.value
            )

    @utils.check_messages(*(STRING_FORMAT_MSGS.keys()))
    def visit_call(self, node):
        func = utils.safe_infer(node.func)
        if isinstance(func, astroid.BoundMethod) and func.name == 'format':
            # If there's a .format() call, run the code below

            if isinstance(node.func.expr, (astroid.Name, astroid.Const)):
                # This is for:
                #   foo = 'Foo {} bar'
                #   print(foo.format(blah)
                for inferred in node.func.expr.infer():
                    if not hasattr(inferred, 'value'):
                        # If there's no value attribute, it's not worth
                        # checking.
                        continue

                    if not isinstance(inferred.value, six.string_types):
                        # If it's not a string, continue
                        continue

                    if '!r}' in inferred.value:
                        self.add_message(
                            'E1322', node=inferred, args=inferred.value
                        )

                    if BAD_FORMATTING_SLOT.findall(inferred.value):
                        if self.config.un_indexed_curly_braces_always_error or \
                                sys.version_info[:2] < (2, 7):
                            self.add_message(
                                'E1320', node=inferred, args=inferred.value
                            )
                        elif six.PY2:
                            self.add_message(
                                'W1320', node=inferred, args=inferred.value
                            )

                try:
                    # Walk back up until no parents are found and look for a
                    # logging.RootLogger instance in the parent types
                    ptr = node
                    while True:
                        parent = ptr.parent
                        for inferred in parent.func.expr.infer():
                            try:
                                instance_type = inferred.pytype().split('.')[0]
                            except TypeError:
                                continue
                            if instance_type == 'logging':
                                self.add_message(
                                    'E1323',
                                    node=node,
                                    args=node.as_string(),
                                )
                                break
                        ptr = parent
                except (AttributeError, InferenceError, NameInferenceError):
                    pass

            elif not hasattr(node.func.expr, 'value'):
                # If it does not have an value attribute, it's not worth
                # checking
                return
            elif isinstance(node.func.expr.value, astroid.Name):
                # No need to check these either
                return
            elif BAD_FORMATTING_SLOT.findall(node.func.expr.value):
                if self.config.un_indexed_curly_braces_always_error or \
                        sys.version_info[:2] < (2, 7):
                    msgid = 'E1320'
                else:
                    msgid = 'W1320'
                self.add_message(
                    'E1320', node=node, args=node.func.expr.value
                )


STRING_LITERALS_MSGS = {
    'E1400': ('Null byte used in unicode string literal (should be wrapped in str())',
              'null-byte-unicode-literal',
              'Null byte used in unicode string literal'),
}


class StringLiteralChecker(BaseTokenChecker):
    '''
    Check string literals
    '''
    __implements__ = (ITokenChecker, IRawChecker)
    name = 'string_literal'
    msgs = STRING_LITERALS_MSGS

    def process_module(self, module):
        self._unicode_literals = 'unicode_literals' in module.future_imports

    def process_tokens(self, tokens):
        for (tok_type, token, (start_row, _), _, _) in tokens:
            if tok_type == tokenize.STRING:
                # 'token' is the whole un-parsed token; we can look at the start
                # of it to see whether it's a raw or unicode string etc.
                self.process_string_token(token, start_row)

    def process_string_token(self, token, start_row):
        if not six.PY2:
            return
        for i, c in enumerate(token):
            if c in '\'\"':
                quote_char = c
                break
        # pylint: disable=undefined-loop-variable
        prefix = token[:i].lower() #  markers like u, b, r.
        after_prefix = token[i:]
        if after_prefix[:3] == after_prefix[-3:] == 3 * quote_char:
            string_body = after_prefix[3:-3]
        else:
            string_body = after_prefix[1:-1]  # Chop off quotes
        # No special checks on raw strings at the moment.
        if 'r' not in prefix:
            self.process_non_raw_string_token(prefix, string_body, start_row)

    def process_non_raw_string_token(self, prefix, string_body, start_row):
        '''
        check for bad escapes in a non-raw string.

        prefix: lowercase string of eg 'ur' string prefix markers.
        string_body: the un-parsed body of the string, not including the quote
        marks.
        start_row: integer line number in the source.
        '''
        if 'u' in prefix:
            if string_body.find('\\0') != -1:
                self.add_message('null-byte-unicode-literal', line=start_row)

def register(linter):
    '''required method to auto register this checker '''
    linter.register_checker(StringCurlyBracesFormatIndexChecker(linter))
    linter.register_checker(StringLiteralChecker(linter))



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Pedro Algarvio (pedro@algarvio.me)`
    :copyright: © 2015 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    ====================================
    PyLint File Permissions Check Plugin
    ====================================

    PyLint plugin which checks for specific file permissions
'''

# Import Python libs
from __future__ import absolute_import
import os
import sys
import glob
import stat

# Import PyLint libs
from pylint.interfaces import IRawChecker
from saltpylint.checkers import BaseChecker


class FilePermsChecker(BaseChecker):
    '''
    Check for files with undesirable permissions
    '''

    __implements__ = IRawChecker

    name = 'fileperms'
    msgs = {'E0599': ('Module file has the wrong file permissions(expected %s): %s',
                      'file-perms',
                      ('Wrong file permissions')),
           }

    priority = -1

    options = (('fileperms-default',
                {'default': '0644', 'type': 'string', 'metavar': 'ZERO_PADDED_PERM',
                 'help': 'Desired file permissons. Default: 0644'}
               ),
               ('fileperms-ignore-paths',
                {'default': (), 'type': 'csv', 'metavar': '<comma-separated-list>',
                 'help': 'File paths to ignore file permission. Glob patterns allowed.'}
               )
              )

    def process_module(self, node):
        '''
        process a module
        '''

        for listing in self.config.fileperms_ignore_paths:
            if node.file.split('{0}/'.format(os.getcwd()))[-1] in glob.glob(listing):
                # File is ignored, no checking should be done
                return

        desired_perm = self.config.fileperms_default
        if '-' in desired_perm:
            desired_perm = desired_perm.split('-')
        else:
            desired_perm = [desired_perm]

        if len(desired_perm) > 2:
            raise RuntimeError('Permission ranges should be like XXXX-YYYY')

        for idx, perm in enumerate(desired_perm):
            desired_perm[idx] = desired_perm[idx].strip('"').strip('\'').lstrip('0').zfill(4)
            if desired_perm[idx][0] != '0':
                # Always include a leading zero
                desired_perm[idx] = '0{0}'.format(desired_perm[idx])
            if sys.version_info > (3,):
                # The octal representation in python 3 has changed to 0o644 instead of 0644
                if desired_perm[idx][1] != 'o':
                    desired_perm[idx] = '0o' + desired_perm[idx][1:]
            if sys.platform.startswith('win'):
                # Windows does not distinguish between user/group/other.
                # They must all be the same. Also, Windows will automatically
                # set the execution bit on files with a known extension
                # (eg .exe, .bat, .com). So we cannot reliably test the
                # execution bit on other files such as .py files.
                user_perm_noexec = int(desired_perm[idx][-3])
                if user_perm_noexec % 2 == 1:
                    user_perm_noexec -= 1
                desired_perm[idx] = desired_perm[idx][:-3] + (str(user_perm_noexec) * 3)

        module_perms = oct(stat.S_IMODE(os.stat(node.file).st_mode))
        if sys.version_info < (3,):
            module_perms = str(module_perms)

        if len(desired_perm) == 1:
            if module_perms != desired_perm[0]:
                if sys.platform.startswith('win'):
                    # Check the variant with execution bit set due to the
                    # unreliability of checking the execution bit on Windows.
                    user_perm_noexec = int(desired_perm[0][-3])
                    desired_perm_exec = desired_perm[0][:-3] + (str(user_perm_noexec + 1) * 3)
                    if module_perms == desired_perm_exec:
                        return
                self.add_message('E0599', line=1, args=(desired_perm[0], module_perms))
        else:
            if module_perms < desired_perm[0] or module_perms > desired_perm[1]:
                if sys.platform.startswith('win'):
                    # Check the variant with execution bit set due to the
                    # unreliability of checking the execution bit on Windows.
                    user_perm_noexec0 = int(desired_perm[0][-3])
                    desired_perm_exec0 = desired_perm[0][:-3] + (str(user_perm_noexec0 + 1) * 3)
                    user_perm_noexec1 = int(desired_perm[1][-3])
                    desired_perm_exec1 = desired_perm[1][:-3] + (str(user_perm_noexec1 + 1) * 3)
                    if desired_perm_exec0 <= module_perms <= desired_perm_exec1:
                        return
                desired_perm = '>= {0} OR <= {1}'.format(*desired_perm)
                self.add_message('E0599', line=1, args=(desired_perm, module_perms))


def register(linter):
    '''
    required method to auto register this checker
    '''
    linter.register_checker(FilePermsChecker(linter))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
The setup script for SaltPyLint
'''
# pylint: disable=file-perms,wrong-import-position
from __future__ import absolute_import, with_statement
import io
import os
import sys

SETUP_KWARGS = {}
USE_SETUPTOOLS = False

# Change to salt source's directory prior to running any command
try:
    SETUP_DIRNAME = os.path.dirname(__file__)
except NameError:
    # We're most likely being frozen and __file__ triggered this NameError
    # Let's work around that
    SETUP_DIRNAME = os.path.dirname(sys.argv[0])


if SETUP_DIRNAME != '':
    os.chdir(SETUP_DIRNAME)

SALT_PYLINT_REQS = os.path.join(os.path.abspath(SETUP_DIRNAME), 'requirements.txt')


def _parse_requirements_file(requirements_file):
    '''
    Parse requirements.txt and return list suitable for
    passing to ``install_requires`` parameter in ``setup()``.
    '''
    parsed_requirements = []
    with open(requirements_file) as rfh:
        for line in rfh.readlines():
            line = line.strip()
            if not line or line.startswith(('#', '-r')):
                continue
            parsed_requirements.append(line)
    return parsed_requirements


def _release_version():
    '''
    Returns release version
    '''
    with io.open(os.path.join(SETUP_DIRNAME, 'saltpylint', 'version.py'), encoding='utf-8') as fh_:
        exec_locals = {}
        exec_globals = {}
        contents = fh_.read()
        if not isinstance(contents, str):
            contents = contents.encode('utf-8')
        exec(contents, exec_globals, exec_locals)  # pylint: disable=exec-used
        return exec_locals['__version__']


# Use setuptools only if the user opts-in by setting the USE_SETUPTOOLS env var.
# Or if setuptools was previously imported (which is the case when using 'pip').
# This ensures consistent behavior, but allows for advanced usage with
# virtualenv, buildout, and others.
if 'USE_SETUPTOOLS' in os.environ or 'setuptools' in sys.modules:
    try:
        from setuptools import setup
        USE_SETUPTOOLS = True
        # This allows correct installation of dependencies with ``pip install``.
        SETUP_KWARGS['install_requires'] = _parse_requirements_file(SALT_PYLINT_REQS)
    except ImportError:
        USE_SETUPTOOLS = False

if USE_SETUPTOOLS is False:
    from distutils.core import setup  # pylint: disable=import-error,no-name-in-module

NAME = 'SaltPyLint'
VERSION = _release_version()
DESCRIPTION = (
    'Required PyLint plugins needed in the several SaltStack projects.'
)

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author='Pedro Algarvio',
    author_email='pedro@algarvio.me',
    url='https://github.com/saltstack/salt-pylint',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX :: Linux',
    ],
    packages=[
        'saltpylint',
        'saltpylint.ext',
        'saltpylint/py3modernize',
        'saltpylint/py3modernize/fixes',
    ],
    **SETUP_KWARGS
)



# -*- coding: utf-8 -*-
'''
    :codeauthor: :email:`Pedro Algarvio (pedro@algarvio.me)`
    :copyright: © 2017 by the SaltStack Team, see AUTHORS for more details.
    :license: Apache 2.0, see LICENSE for more details.


    saltpylint.blacklist
    ~~~~~~~~~~~~~~~~~~~~

    Checks blacklisted imports and code usage on salt
'''

# Import python libs
from __future__ import absolute_import
import os
import fnmatch

# Import pylint libs
import astroid
from saltpylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker

BLACKLISTED_IMPORTS_MSGS = {
    'E9402': ('Uses of a blacklisted module %r: %s',
              'blacklisted-module',
              'Used a module marked as blacklisted is imported.'),
    'E9403': ('Uses of a blacklisted external module %r: %s',
              'blacklisted-external-module',
              'Used a module marked as blacklisted is imported.'),
    'E9404': ('Uses of a blacklisted import %r: %s',
              'blacklisted-import',
              'Used an import marked as blacklisted.'),
    'E9405': ('Uses of an external blacklisted import %r: %s',
              'blacklisted-external-import',
              'Used an external import marked as blacklisted.'),
    'E9406': ('Uses of blacklisted test module execution code: %s',
              'blacklisted-test-module-execution',
              'Uses of blacklisted test module execution code.'),
    'E9407': ('Uses of blacklisted sys.path updating through \'ensure_in_syspath\'. '
              'Please remove the import and any calls to \'ensure_in_syspath()\'.',
              'blacklisted-syspath-update',
              'Uses of blacklisted sys.path updating through ensure_in_syspath.'),
    }


class BlacklistedImportsChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'blacklisted-imports'
    msgs = BLACKLISTED_IMPORTS_MSGS
    priority = -2

    def open(self):
        self.blacklisted_modules = ('salttesting',
                                    'integration',
                                    'unit',
                                    'mock',
                                    'six',
                                    'distutils.version',
                                    'unittest',
                                    'unittest2')

    @utils.check_messages('blacklisted-imports')
    def visit_import(self, node):
        '''triggered when an import statement is seen'''
        module_filename = node.root().file
        if fnmatch.fnmatch(module_filename, '__init__.py*') and \
                not fnmatch.fnmatch(module_filename, 'test_*.py*'):
            return
        modnode = node.root()
        names = [name for name, _ in node.names]

        for name in names:
            self._check_blacklisted_module(node, name)

    @utils.check_messages('blacklisted-imports')
    def visit_importfrom(self, node):
        '''triggered when a from statement is seen'''
        module_filename = node.root().file
        if fnmatch.fnmatch(module_filename, '__init__.py*') and \
                not fnmatch.fnmatch(module_filename, 'test_*.py*'):
            return
        basename = node.modname
        self._check_blacklisted_module(node, basename)

    def _check_blacklisted_module(self, node, mod_path):
        '''check if the module is blacklisted'''
        for mod_name in self.blacklisted_modules:
            if mod_path == mod_name or mod_path.startswith(mod_name + '.'):
                names = []
                for name, name_as in node.names:
                    if name_as:
                        names.append('{0} as {1}'.format(name, name_as))
                    else:
                        names.append(name)
                try:
                    import_from_module = node.modname
                    if import_from_module == 'salttesting.helpers':
                        for name in names:
                            if name == 'ensure_in_syspath':
                                self.add_message('blacklisted-syspath-update', node=node)
                                continue
                            msg = 'Please use \'from tests.support.helpers import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module in ('salttesting.mock', 'mock', 'unittest.mock', 'unittest2.mock'):
                        for name in names:
                            msg = 'Please use \'from tests.support.mock import {0}\''.format(name)
                            if import_from_module in ('salttesting.mock', 'unittest.mock', 'unittest2.mock'):
                                message_id = 'blacklisted-module'
                            else:
                                message_id = 'blacklisted-external-module'
                            self.add_message(message_id, node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == 'salttesting.parser':
                        for name in names:
                            msg = 'Please use \'from tests.support.parser import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == 'salttesting.case':
                        for name in names:
                            msg = 'Please use \'from tests.support.case import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == 'salttesting.unit':
                        for name in names:
                            msg = 'Please use \'from tests.support.unit import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module.startswith(('unittest', 'unittest2')):
                        for name in names:
                            msg = 'Please use \'from tests.support.unit import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == 'salttesting.mixins':
                        for name in names:
                            msg = 'Please use \'from tests.support.mixins import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == 'six':
                        for name in names:
                            msg = 'Please use \'from salt.ext.six import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if import_from_module == 'distutils.version':
                        for name in names:
                            msg = 'Please use \'from salt.utils.versions import {0}\''.format(name)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                        continue
                    if names:
                        for name in names:
                            if name in ('TestLoader', 'TextTestRunner', 'TestCase', 'expectedFailure',
                                        'TestSuite', 'skipIf', 'TestResult'):
                                msg = 'Please use \'from tests.support.unit import {0}\''.format(name)
                                self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                                continue
                            if name in ('SaltReturnAssertsMixin', 'SaltMinionEventAssertsMixin'):
                                msg = 'Please use \'from tests.support.mixins import {0}\''.format(name)
                                self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                                continue
                            if name in ('ModuleCase', 'SyndicCase', 'ShellCase', 'SSHCase'):
                                msg = 'Please use \'from tests.support.case import {0}\''.format(name)
                                self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                                continue
                            if name == 'run_tests':
                                msg = 'Please remove the \'if __name__ == "__main__":\' section from the end of the module'
                                self.add_message('blacklisted-test-module-execution', node=node, args=msg)
                                continue
                            if mod_name in ('integration', 'unit'):
                                if name in ('SYS_TMP_DIR',
                                            'TMP',
                                            'FILES',
                                            'PYEXEC',
                                            'MOCKBIN',
                                            'SCRIPT_DIR',
                                            'TMP_STATE_TREE',
                                            'TMP_PRODENV_STATE_TREE',
                                            'TMP_CONF_DIR',
                                            'TMP_SUB_MINION_CONF_DIR',
                                            'TMP_SYNDIC_MINION_CONF_DIR',
                                            'TMP_SYNDIC_MASTER_CONF_DIR',
                                            'CODE_DIR',
                                            'TESTS_DIR',
                                            'CONF_DIR',
                                            'PILLAR_DIR',
                                            'TMP_SCRIPT_DIR',
                                            'ENGINES_DIR',
                                            'LOG_HANDLERS_DIR',
                                            'INTEGRATION_TEST_DIR'):
                                    msg = 'Please use \'from tests.support.paths import {0}\''.format(name)
                                    self.add_message('blacklisted-import', node=node, args=(mod_path, msg))
                                    continue
                                msg = 'Please use \'from tests.{0} import {1}\''.format(mod_path, name)
                                self.add_message('blacklisted-import', node=node, args=(mod_path, msg))
                                continue
                            msg = 'Please report this error to SaltStack so we can fix it: Trying to import {0} from {1}'.format(name, mod_path)
                            self.add_message('blacklisted-module', node=node, args=(mod_path, msg))
                except AttributeError:
                    if mod_name in ('integration', 'unit', 'mock', 'six', 'distutils.version',
                                    'unittest', 'unittest2'):
                        if mod_name in ('integration', 'unit'):
                            msg = 'Please use \'import tests.{0} as {0}\''.format(mod_name)
                            message_id = 'blacklisted-import'
                        elif mod_name == 'mock':
                            msg = 'Please use \'import tests.support.{0} as {0}\''.format(mod_name)
                            message_id = 'blacklisted-external-import'
                        elif mod_name == 'six':
                            msg = 'Please use \'import salt.ext.{0} as {0}\''.format(name)
                            message_id = 'blacklisted-external-import'
                        elif mod_name == 'distutils.version':
                            msg = 'Please use \'import salt.utils.versions\' instead'
                            message_id = 'blacklisted-import'
                        elif mod_name.startswith(('unittest', 'unittest2')):
                            msg = 'Please use \'import tests.support.unit as {}\' instead'.format(mod_name)
                            message_id = 'blacklisted-import'
                        self.add_message(message_id, node=node, args=(mod_path, msg))
                        continue
                    msg = 'Please report this error to SaltStack so we can fix it: Trying to import {0}'.format(mod_path)
                    self.add_message('blacklisted-import', node=node, args=(mod_path, msg))

BLACKLISTED_LOADER_USAGE_MSGS = {
    'E9501': ('Blacklisted salt loader dunder usage. Setting dunder attribute %r to module %r. '
              'Use \'salt.support.mock\' and \'patch.dict()\' instead.',
              'unmocked-patch-dunder',
              'Uses a blacklisted salt loader dunder usage in tests.'),
    'E9502': ('Blacklisted salt loader dunder usage. Setting attribute %r to module %r. '
              'Use \'salt.support.mock\' and \'patch()\' instead.',
              'unmocked-patch',
              'Uses a blacklisted salt loader dunder usage in tests.'),
    'E9503': ('Blacklisted salt loader dunder usage. Updating dunder attribute %r on module %r. '
              'Use \'salt.support.mock\' and \'patch.dict()\' instead.',
              'unmocked-patch-dunder-update',
              'Uses a blacklisted salt loader dunder usage in tests.'),
}


class BlacklistedLoaderModulesUsageChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'blacklisted-unmocked-patching'
    msgs = BLACKLISTED_LOADER_USAGE_MSGS
    priority = -2

    def open(self):
        self.process_module = False
        self.salt_dunders = (
            '__opts__', '__salt__', '__runner__', '__context__', '__utils__',
            '__ext_pillar__', '__thorium__', '__states__', '__serializers__',
            '__ret__', '__grains__', '__pillar__', '__sdb__', '__proxy__',
            '__low__', '__orchestration_jid__', '__running__', '__intance_id__',
            '__lowstate__', '__env__'
        )
        self.imported_salt_modules = {}

    def close(self):
        self.process_module = False
        self.imported_salt_modules = {}

    @utils.check_messages('blacklisted-unmocked-patching')
    def visit_module(self, node):
        module_filename = node.root().file
        if not fnmatch.fnmatch(os.path.basename(module_filename), 'test_*.py*'):
            return
        self.process_module = True

    @utils.check_messages('blacklisted-unmocked-patching')
    def leave_module(self, node):
        if self.process_module:
            # Reset
            self.process_module = False
            self.imported_salt_modules = {}

    @utils.check_messages('blacklisted-unmocked-patching')
    def visit_import(self, node):
        '''triggered when an import statement is seen'''
        if self.process_module:
            # Store salt imported modules
            for module, import_as in node.names:
                if not module.startswith('salt'):
                    continue
                if import_as and import_as not in self.imported_salt_modules:
                    self.imported_salt_modules[import_as] = module
                    continue
                if module not in self.imported_salt_modules:
                    self.imported_salt_modules[module] = module

    @utils.check_messages('blacklisted-unmocked-patching')
    def visit_importfrom(self, node):
        '''triggered when a from statement is seen'''
        if self.process_module:
            if not node.modname.startswith('salt'):
                return
            # Store salt imported modules
            for module, import_as in node.names:
                if import_as and import_as not in self.imported_salt_modules:
                    self.imported_salt_modules[import_as] = import_as
                    continue
                if module not in self.imported_salt_modules:
                    self.imported_salt_modules[module] = module

    @utils.check_messages('blacklisted-loader-usage')
    def visit_assign(self, node, *args):
        if not self.process_module:
            return

        node_left = node.targets[0]

        if isinstance(node_left, astroid.Subscript):
            # Were're changing an existing attribute
            if not isinstance(node_left.value, astroid.Attribute):
                return
            if node_left.value.attrname in self.salt_dunders:
                self.add_message(
                    'unmocked-patch-dunder-update',
                    node=node,
                    args=(node_left.value.attrname,
                          self.imported_salt_modules[node_left.value.expr.name])
                )
                return

        if not isinstance(node_left, astroid.AssignAttr):
            return

        try:
            if node_left.expr.name not in self.imported_salt_modules:
                # If attributes are not being set on salt's modules,
                # leave it alone, for now!
                return
        except AttributeError:
            # This mmight not be what we're looking for
            return

        # we're assigning to an imported salt module!
        if node_left.attrname in self.salt_dunders:
            # We're changing salt dunders
            self.add_message(
                'unmocked-patch-dunder',
                node=node,
                args=(node_left.attrname,
                      self.imported_salt_modules[node_left.expr.name])
            )
            return

        # Changing random attributes
        self.add_message(
            'unmocked-patch',
            node=node,
            args=(node_left.attrname,
                  self.imported_salt_modules[node_left.expr.name])
        )


RESOURCE_LEAKAGE_MSGS = {
    'W8470': ('Resource leakage detected. %s ',
              'resource-leakage',
              'Resource leakage detected.'),
}


class ResourceLeakageChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'resource-leakage'
    msgs = RESOURCE_LEAKAGE_MSGS
    priority = -2

    def open(self):
        self.inside_with_ctx = False

    def close(self):
        self.inside_with_ctx = False

    def visit_with(self, node):
        self.inside_with_ctx = True

    def leave_with(self, node):
        self.inside_with_ctx = False

    def visit_call(self, node):
        if isinstance(node.func, astroid.Attribute):
            if node.func.attrname == 'fopen' and self.inside_with_ctx is False:
                msg = ('Please call \'salt.utils.fopen\' using the \'with\' context '
                       'manager, otherwise the file handle won\'t be closed and '
                       'resource leakage will occur.')
                self.add_message('resource-leakage', node=node, args=(msg,))
        elif isinstance(node.func, astroid.Name):
            if utils.is_builtin(node.func.name) and node.func.name == 'open':
                if self.inside_with_ctx:
                    msg = ('Please use \'with salt.utils.fopen()\' instead of '
                           '\'with open()\'. It assures salt does not leak '
                           'file handles.')
                else:
                    msg = ('Please use \'salt.utils.fopen()\' instead of \'open()\' '
                           'using the \'with\' context manager, otherwise the file '
                           'handle won\'t be closed and resource leakage will occur.')
                self.add_message('resource-leakage', node=node, args=(msg,))


MOVED_TEST_CASE_CLASSES_MSGS = {
    'E9490': ('Moved test case base class detected. %s',
              'moved-test-case-class',
              'Moved test case base class detected.'),
    'E9491': ('Moved test case mixin class detected. %s',
              'moved-test-case-mixin',
              'Moved test case mixin class detected.'),
}


class MovedTestCaseClassChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'moved-test-case-class'
    msgs = MOVED_TEST_CASE_CLASSES_MSGS
    priority = -2

    def open(self):
        self.process_module = False

    def close(self):
        self.process_module = False

    @utils.check_messages('moved-test-case-class')
    def visit_module(self, node):
        module_filename = node.root().file
        if not fnmatch.fnmatch(os.path.basename(module_filename), 'test_*.py*'):
            return
        self.process_module = True

    @utils.check_messages('moved-test-case-class')
    def leave_module(self, node):
        if self.process_module:
            # Reset
            self.process_module = False

    @utils.check_messages('moved-test-case-class')
    def visit_importfrom(self, node):
        '''triggered when a from statement is seen'''
        if self.process_module:
            if not node.modname.startswith('tests.integration'):
                return
            # Store salt imported modules
            for module, import_as in node.names:
                if import_as:
                    self._check_moved_imports(node, module, import_as)
                    continue
                self._check_moved_imports(node, module)

    @utils.check_messages('moved-test-case-class')
    def visit_classdef(self, node):
        for base in node.bases:
            if not hasattr(base, 'attrname'):
                continue
            if base.attrname in ('TestCase',):
                msg = 'Please use \'from tests.support.unit import {0}\''.format(base.attrname)
                self.add_message('moved-test-case-class', node=node, args=(msg,))
            if base.attrname in ('ModuleCase', 'SyndicCase', 'ShellCase', 'SSHCase'):
                msg = 'Please use \'from tests.support.case import {0}\''.format(base.attrname)
                self.add_message('moved-test-case-class', node=node, args=(msg,))
            if base.attrname in ('AdaptedConfigurationTestCaseMixin', 'ShellCaseCommonTestsMixin',
                                 'SaltMinionEventAssertsMixin'):
                msg = 'Please use \'from tests.support.mixins import {0}\''.format(base.attrname)
                self.add_message('moved-test-case-mixin', node=node, args=(msg,))

    def _check_moved_imports(self, node, module, import_as=None):
        names = []
        for name, name_as in node.names:
            if name not in ('ModuleCase', 'SyndicCase', 'ShellCase', 'SSHCase'):
                continue
            if name_as:
                msg = 'Please use \'from tests.support.case import {0} as {1}\''.format(name, name_as)
            else:
                msg = 'Please use \'from tests.support.case import {0}\''.format(name)
            self.add_message('moved-test-case-class', node=node, args=(msg,))


BLACKLISTED_FUNCTIONS_MSGS = {
    'E9601': ('Use of blacklisted function %s (use %s instead)',
              'blacklisted-function',
              'Used a function marked as blacklisted'),
    }


class BlacklistedFunctionsChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = 'blacklisted-functions'
    msgs = BLACKLISTED_FUNCTIONS_MSGS
    priority = -2
    max_depth = 20

    options = (
        ('blacklisted-functions',
         {'default': '', 'type': 'string',
          'metavar': 'bad1=good1,bad2=good2',
          'help': 'List of blacklisted functions and their recommended '
                  'replacements'}),
    )

    def open(self):
        self.blacklisted_functions = {}
        blacklist = [
            x.strip() for x in self.config.blacklisted_functions.split(',')]
        for item in [x.strip() for x in
                     self.config.blacklisted_functions.split(',')]:
            try:
                key, val = [x.strip() for x in item.split('=')]
            except ValueError:
                pass
            else:
                self.blacklisted_functions[key] = val

    def _get_full_name(self, node):
        try:
            func = utils.safe_infer(node.func)
            if func.name.__str__() == 'Uninferable':
                return
        except Exception:
            func = None
        if func is None:
            return
        ret = []
        depth = 0
        while func is not None:
            depth += 1
            if depth > self.max_depth:
                # Prevent endless loop
                return
            try:
                ret.append(func.name)
            except AttributeError:
                return
            func = func.parent
        # ret will contain the levels of the function from last to first (e.g.
        # ['walk', 'os']. Reverse it and join with dots to get the correct
        # full name for the function.
        return '.'.join(ret[::-1])

    @utils.check_messages('blacklisted-functions')
    def visit_call(self, node):
        if self.blacklisted_functions:
            full_name = self._get_full_name(node)
            if full_name is not None:
                try:
                    self.add_message(
                        'blacklisted-function',
                        node=node,
                        args=(full_name, self.blacklisted_functions[full_name])
                    )
                except KeyError:
                    # Not blacklisted
                    pass


def register(linter):
    '''
    Required method to auto register this checker
    '''
    linter.register_checker(ResourceLeakageChecker(linter))
    linter.register_checker(BlacklistedImportsChecker(linter))
    linter.register_checker(MovedTestCaseClassChecker(linter))
    linter.register_checker(BlacklistedLoaderModulesUsageChecker(linter))
    linter.register_checker(BlacklistedFunctionsChecker(linter))
